@@ -1,4 +1,4 @@
-#     Copyright 2020, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -32,14 +32,17 @@ Values can be seen as:
 * LoopComplete (complete knowledge of loop types)
 """
 
-
 from nuitka.nodes.shapes.StandardShapes import (
     ShapeLoopCompleteAlternative,
     ShapeLoopInitialAlternative,
     tshape_uninit,
     tshape_unknown,
 )
-from nuitka.utils import InstanceCounters
+from nuitka.utils.InstanceCounters import (
+    counted_del,
+    counted_init,
+    isCountingInstances,
+)
 
 
 class ValueTraceBase(object):
@@ -55,7 +58,7 @@ class ValueTraceBase(object):
         "previous",
     )
 
-    @InstanceCounters.counted_init
+    @counted_init
     def __init__(self, owner, previous):
         self.owner = owner
 
@@ -73,7 +76,8 @@ class ValueTraceBase(object):
         # Previous trace this is replacing.
         self.previous = previous
 
-    __del__ = InstanceCounters.counted_del()
+    if isCountingInstances():
+        __del__ = counted_del()
 
     def __repr__(self):
         return "<%s of %s>" % (self.__class__.__name__, self.owner.getCodeName())
@@ -161,10 +165,14 @@ class ValueTraceBase(object):
 
         return None
 
-    def hasShapeDictionaryExact(self):
-        # Virtual method, pylint: disable=no-self-use
+    @staticmethod
+    def hasShapeDictionaryExact():
         return False
 
+    @staticmethod
+    def getTruthValue():
+        return None
+
 
 class ValueTraceUnassignedBase(ValueTraceBase):
     __slots__ = ()
@@ -354,6 +362,9 @@ class ValueTraceAssign(ValueTraceBase):
     def hasShapeDictionaryExact(self):
         return self.assign_node.subnode_source.hasShapeDictionaryExact()
 
+    def getTruthValue(self):
+        return self.assign_node.subnode_source.getTruthValue()
+
 
 class ValueTraceMergeBase(ValueTraceBase):
     """Merge of two or more traces or start of loops."""
@@ -440,11 +451,31 @@ class ValueTraceMerge(ValueTraceMergeBase):
         self.usage_count += 1
 
     def hasShapeDictionaryExact(self):
-        for previous in self.previous:
-            if not previous.hasShapeDictionaryExact():
-                return False
+        return all(previous.hasShapeDictionaryExact() for previous in self.previous)
 
-        return True
+    def getTruthValue(self):
+        any_false = False
+        any_true = False
+
+        for previous in self.previous:
+            truth_value = previous.getTruthValue()
+
+            # One unknown kills it.
+            if truth_value is None:
+                return None
+            elif truth_value is True:
+                # True and false values resembled unknown.
+                if any_false:
+                    return None
+                any_true = True
+            else:
+                # True and false values resembled unknown.
+                if any_true:
+                    return None
+                any_false = True
+
+        # Now all agreed and were not unknown, so we can conclude all false or all true.
+        return any_true
 
 
 class ValueTraceLoopBase(ValueTraceMergeBase):
@@ -515,6 +546,7 @@ class ValueTraceLoopComplete(ValueTraceLoopBase):
             and self.type_shapes == other.type_shapes
         )
 
+    # TODO: These could be better
     @staticmethod
     def mustHaveValue():
         return False
@@ -523,6 +555,10 @@ class ValueTraceLoopComplete(ValueTraceLoopBase):
     def mustNotHaveValue():
         return False
 
+    @staticmethod
+    def getTruthValue():
+        return None
+
 
 class ValueTraceLoopIncomplete(ValueTraceLoopBase):
     __slots__ = ()
@@ -544,3 +580,7 @@ class ValueTraceLoopIncomplete(ValueTraceLoopBase):
     @staticmethod
     def mustNotHaveValue():
         return False
+
+    @staticmethod
+    def getTruthValue():
+        return None
